/*  $Header: /home/cvsroot/dvipdfmx/src/pdfdraw.c,v 1.2 2004/08/07 09:52:34 hirata Exp $
    
    This is dvipdfmx, an eXtended version of dvipdfm by Mark A. Wicks.

    Copyright (C) 2002 by Jin-Hwan Cho and Shunsaku Hirata,
    the dvipdfmx project team <dvipdfmx@project.ktug.or.kr>
    
    Copyright (C) 1998, 1999 by Mark A. Wicks <mwicks@kettering.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <math.h>

#include "system.h"
#include "error.h"
#include "mem.h"
#include "mfileio.h"

#include "numbers.h"

#include "pdfdoc.h"
#include "pdfdev.h"

#include "pdfdraw.h"

struct path_element
{
  char type;
  pdf_coord p[3];
};

struct pdf_path
{
  int num_paths;
  int max_paths;
  struct path_element *path;
};

pdf_path *
pdf_draw_new_path (void)
{
  pdf_path *p;

  p = NEW(1, pdf_path);
  p->num_paths = 0;
  p->max_paths = 0;
  p->path  = NULL;

  return p;
}

void
pdf_draw_setmiterlimit (double mlimit)
{
  int len = 0;

  work_buffer[len++] = ' ';
  len += pdf_sprint_length(work_buffer+len, mlimit);
  work_buffer[len++] = ' ';
  work_buffer[len++] = 'M';
  pdf_doc_add_to_page(work_buffer, len);
}

void
pdf_draw_setlinecap (int capstyle)
{
  int len;

  len = sprintf(work_buffer, " %d J", capstyle);
  pdf_doc_add_to_page(work_buffer, len);
}

void
pdf_draw_setlinejoin (int joinstyle)
{
  int len;

  len = sprintf(work_buffer, " %d j", joinstyle);
  pdf_doc_add_to_page(work_buffer, len);
}

void
pdf_draw_setlinewidth (double width)
{
  int len = 0;

  work_buffer[len++] = ' ';
  len += pdf_sprint_length(work_buffer+len, width);
  work_buffer[len++] = ' ';
  work_buffer[len++] = 'w';
  pdf_doc_add_to_page(work_buffer, len);
}

void
pdf_draw_setdash (int count, double *pattern, double offset)
{
  int i, len = 0;

  pdf_doc_add_to_page(" [", 2);
  for (i = 0; i < count; i++) {
    work_buffer[0] = ' ';
    len = pdf_sprint_length(work_buffer+1, pattern[i]);
    pdf_doc_add_to_page(work_buffer, len+1);
  }
  pdf_doc_add_to_page("] ", 2);
  len = pdf_sprint_length(work_buffer, offset);
  pdf_doc_add_to_page(work_buffer, len);
  pdf_doc_add_to_page(" d", 2);
}


/*
 *  x y width height re
 * 
 * is equivalent to
 *
 *  x y m
 *  (x + width) y l
 *  (x + width) (y + height) l
 *  x (y + height) l
 *  h
 */
static int
is_rect (const pdf_path *p, int ignore_rule)
{
  struct path_element *pe0, *pe1, *pe2, *pe3, *pe4;

  if (p->num_paths == 5) {
    pe0 = &(p->path[0]);
    pe1 = &(p->path[1]);
    pe2 = &(p->path[2]);
    pe3 = &(p->path[3]);
    pe4 = &(p->path[4]);
    if (pe0->type == 'm' &&
	pe1->type == 'l' &&
	pe2->type == 'l' &&
	pe3->type == 'l' &&
	pe4->type == 'h') {
      if (pe1->p[0].y - pe0->p[0].y == 0 &&
	  pe2->p[0].x - pe1->p[0].x == 0 &&
	  pe3->p[0].y - pe2->p[0].y == 0) {
	if (pe1->p[0].x - pe0->p[0].x
	    == pe2->p[0].x - pe3->p[0].x) {
	  return 1;
	}
	/* Winding number is different but ignore it here. */
      } else if (ignore_rule &&
		 pe1->p[0].x - pe0->p[0].x == 0 &&
		 pe2->p[0].y - pe1->p[0].y == 0 &&
		 pe3->p[0].x - pe2->p[0].x == 0) {
	if (pe1->p[0].y - pe0->p[0].y
	    == pe2->p[0].y - pe3->p[0].y) {
	  return 1;
	}
      }
    }
  }

  return 0;
}

void
pdf_draw_path (const pdf_path *p, const char *p_op)
{
  struct path_element *pe, *pe1;
  int i, len = 0;

  if (is_rect(p, 1)) {
    pdf_rect rect; /* FIXME */

    pe  = &(p->path[0]);
    pe1 = &(p->path[2]);

    rect.llx = pe->p[0].x;
    rect.lly = pe->p[0].y;
    rect.urx = pe1->p[0].x - pe->p[0].x; /* width...  */
    rect.ury = pe1->p[0].y - pe->p[0].y; /* height... */

    work_buffer[len++] = ' ';
    len += pdf_sprint_rect(work_buffer+len, &rect);
    work_buffer[len++] = ' ';
    work_buffer[len++] = 'r';
    work_buffer[len++] = 'e';
    pdf_doc_add_to_page(work_buffer, len);
  } else {
    for (i = 0; i < p->num_paths; i++) {
      len = 0;
      pe = &(p->path[i]);
      switch (pe->type) {
      case 'm': case 'l':
	work_buffer[len++] = ' ';
	len += pdf_sprint_coord(work_buffer+len, &pe->p[0]);
	work_buffer[len++] = ' ';
	work_buffer[len++] = pe->type;
	break;
      case 'v':	/* First control point coincide with path's start point. */
      case 'y':	/* Second control point coincide with path's end point.  */
	work_buffer[len++] = ' ';
	len += pdf_sprint_coord(work_buffer+len, &pe->p[0]);
	work_buffer[len++] = ' ';
	len += pdf_sprint_coord(work_buffer+len, &pe->p[1]);
	work_buffer[len++] = ' ';
	work_buffer[len++] = pe->type;
	break;
      case 'c':
	work_buffer[len++] = ' ';
	len += pdf_sprint_coord(work_buffer+len, &pe->p[0]);
	work_buffer[len++] = ' ';
	len += pdf_sprint_coord(work_buffer+len, &pe->p[1]);
	work_buffer[len++] = ' ';
	len += pdf_sprint_coord(work_buffer+len, &pe->p[2]);
	work_buffer[len++] = ' ';
	work_buffer[len++] = pe->type;
	break;
      case 'h': /* Closepath */
	work_buffer[len++] = ' ';
	work_buffer[len++] = 'h';
	break;
      default:
	ERROR("Unknown path type: %d", pe->type);
      }
      pdf_doc_add_to_page(work_buffer, len);
    }
  }

  pdf_doc_add_to_page(" ", 1);
  pdf_doc_add_to_page(p_op, strlen(p_op));
}

void
pdf_draw_clear_path (pdf_path *p)
{
  if (p)
    p->num_paths = 0;
}

void
pdf_draw_release_path (pdf_path *p)
{
  if (p) {
    if (p->path)
      RELEASE(p->path);
    p->path = NULL;
    p->num_paths = 0;
    p->max_paths = 0;
  }
}

int
pdf_draw_copy_path (pdf_path *dst, pdf_path *src)
{
  int i;

  dst->num_paths = src->num_paths;

  if (dst->max_paths < src->num_paths) {
    dst->path = RENEW(dst->path, src->num_paths, struct path_element);
    dst->max_paths = src->num_paths;
  }

  for (i = 0; i < src->num_paths; i++) {
    struct path_element *pe1, *pe2;

    pe1 = &(dst->path[i]);
    pe2 = &(src->path[i]);
    pe1->type = pe2->type;
    pe1->p[0].x = pe2->p[0].x; pe1->p[0].y = pe2->p[0].y;
    pe1->p[1].x = pe2->p[1].x; pe1->p[1].y = pe2->p[1].y;
    pe1->p[2].x = pe2->p[2].x; pe1->p[2].y = pe2->p[2].y;
  }

  return 0;
}


#define PATH_ALLOC_SIZE 64

/*
 * cp (currentpoint) will be modified.
 */
int
pdf_draw_append_path (pdf_path *p, pdf_coord *cp, int type, double *points)
{
  struct path_element *pe, *pe0;

  if (!p || !cp)
    return 1;

  if (p->num_paths >= p->max_paths) {
    p->max_paths += PATH_ALLOC_SIZE;
    p->path = RENEW(p->path, p->max_paths, struct path_element);
  }

  if (p->num_paths > 0)
    pe0 = &(p->path[p->num_paths-1]);
  else
    pe0 = NULL;
  pe = &(p->path[p->num_paths]);

  switch (type) {
  case PDF_PATH_MOVETO:
    if (pe0 && pe0->type == 'm') {
      pe0->p[0].x = points[0];
      pe0->p[0].y = points[1];
      cp->x = pe0->p[0].x;
      cp->y = pe0->p[0].y;
    } else {
      pe->type = 'm';
      pe->p[0].x = points[0];
      pe->p[0].y = points[1];
      cp->x = pe->p[0].x;
      cp->y = pe->p[0].y;
      p->num_paths++;
    }
    break;
  case PDF_PATH_RMOVETO:
    if (pe0 && pe0->type == 'm') {
      pe0->p[0].x = cp->x + points[0];
      pe0->p[0].y = cp->y + points[1];
      cp->x = pe0->p[0].x;
      cp->y = pe0->p[0].y;
    } else {
      pe->type = 'm';
      pe->p[0].x = points[0] + cp->x;
      pe->p[0].y = points[1] + cp->y;
      cp->x = pe->p[0].x;
      cp->y = pe->p[0].y;
      p->num_paths++;
    }
    break;
  case PDF_PATH_LINETO:
    pe->type = 'l';
    pe->p[0].x = points[0];
    pe->p[0].y = points[1];
    cp->x = pe->p[0].x;
    cp->y = pe->p[0].y;
    p->num_paths++;
    break;
  case PDF_PATH_RLINETO:
    pe->type = 'l';
    pe->p[0].x = points[0] + cp->x;
    pe->p[0].y = points[1] + cp->y;
    cp->x = pe->p[0].x;
    cp->y = pe->p[0].y;
    p->num_paths++;
    break;
  case PDF_PATH_CURVETO:
    if (points[0] == cp->x &&
	points[1] == cp->y) {
      pe->type = 'v';
      pe->p[0].x = points[2];
      pe->p[0].y = points[3];
      pe->p[1].x = points[4];
      pe->p[1].y = points[5];
      cp->x = pe->p[1].x;
      cp->y = pe->p[1].y;
    } else if (points[2] == points[4] &&
	       points[3] == points[5]) {
      pe->type = 'y';
      pe->p[0].x = points[0];
      pe->p[0].y = points[1];
      pe->p[1].x = points[2];
      pe->p[1].y = points[3];
      cp->x = pe->p[1].x;
      cp->y = pe->p[1].y;
    } else {
      pe->type = 'c';
      pe->p[0].x = points[0];
      pe->p[0].y = points[1];
      pe->p[1].x = points[2];
      pe->p[1].y = points[3];
      pe->p[2].x = points[4];
      pe->p[2].y = points[5];
      cp->x = pe->p[2].x;
      cp->y = pe->p[2].y;
    }
    p->num_paths++;
    break;
  case PDF_PATH_RCURVETO:
    if (points[0] == 0.0 && points[1] == 0.0) {
      pe->type = 'v';
      pe->p[0].x = points[2] + cp->x;
      pe->p[0].y = points[3] + cp->y;
      pe->p[1].x = points[4] + cp->x;
      pe->p[1].y = points[5] + cp->y;
      cp->x = pe->p[1].x;
      cp->y = pe->p[1].y;
    } else if (points[2] == points[4] &&
	       points[3] == points[5]) {
      pe->type = 'y';
      pe->p[0].x = points[0] + cp->x;
      pe->p[0].y = points[1] + cp->y;
      pe->p[1].x = points[2] + cp->x;
      pe->p[1].y = points[3] + cp->y;
      cp->x = pe->p[1].x;
      cp->y = pe->p[1].y;
    } else {
      pe->type = 'c';
      pe->p[0].x = points[0] + cp->x;
      pe->p[0].y = points[1] + cp->y;
      pe->p[1].x = points[2] + cp->x;
      pe->p[1].y = points[3] + cp->y;
      pe->p[2].x = points[4] + cp->x;
      pe->p[2].y = points[5] + cp->y;
      cp->x = pe->p[2].x;
      cp->y = pe->p[2].y;
    }
    p->num_paths++;
    break;
  case PDF_PATH_CLOSE:
    if (p->num_paths > 0) {
      pe0 = &(p->path[0]);
      if ((cp->x != pe0->p[0].x ||
	   cp->y != pe0->p[0].y)) {
	pe->type = 'h';
	cp->x = pe0->p[2].x;
	cp->y = pe0->p[2].y;
	p->num_paths++;
      }
    }
    break;

  default:
    WARN("Unknown path type: %d", type);
    return 1;
  }

  return 0;
}

/* Modify M itself */
void
pdf_invertmatrix (pdf_tmatrix *M)
{
  pdf_tmatrix M_i;  
  double det;

  det = (M->a) * (M->d) - (M->b) * (M->c);
  if (det == 0.0)
    ERROR("Determinant exactly zero in pdf_invertmatrix()");

  M_i.a =  (M->d) / det; M_i.b = -(M->b) / det;
  M_i.c = -(M->c) / det; M_i.d =  (M->a) / det;
  M_i.e = ((M->b) * (M->f) - (M->d) * (M->e)) / det;
  M_i.f = ((M->e) * (M->c) - (M->a) * (M->f)) / det;

  pdf_copymatrix(M, &M_i);
}

void
pdf_draw_dtransform_coord (pdf_coord *p, const pdf_tmatrix *M)
{
  double x, y;

  x = p->x; y = p->y;
  if (M) {
    p->x = x * M->a + y * M->c;
    p->y = x * M->b + y * M->d;
  }
}

void
pdf_draw_idtransform_coord (pdf_coord *p, const pdf_tmatrix *M)
{
  pdf_tmatrix M_i;

  pdf_copymatrix(&M_i, M);
  pdf_invertmatrix(&M_i);
  pdf_draw_dtransform_coord(p, &M_i);
}

void
pdf_draw_transform_coord (pdf_coord *p, const pdf_tmatrix *M)
{
  double x, y;

  x = p->x; y = p->y;
  p->x = x * M->a + y * M->c + M->e;
  p->y = x * M->b + y * M->d + M->f;
}

void
pdf_draw_itransform_coord (pdf_coord *p, const pdf_tmatrix *M)
{
  pdf_tmatrix M_i;

  pdf_copymatrix(&M_i, M);
  pdf_invertmatrix(&M_i);
  pdf_draw_transform_coord(p, &M_i);
}

void
pdf_draw_transform_path (pdf_path *p, const pdf_tmatrix *M)
{
  struct path_element *pe;
  int n = 0, i;

  for (i = 0; i < p->num_paths; i++) {
    pe = &(p->path[i]);
    switch (pe->type) {
    case 'm': n = 1; break;
    case 'l': n = 1; break;
    case 'c': n = 3; break;
    case 'v': n = 2; break;
    case 'y': n = 2; break;
    case 'h': n = 0; break;
    default:
      ERROR("Unknown path type: %d", pe->type);
    }
    while (n-- > 0) {
      pdf_draw_transform_coord(&(pe->p[n]), M);
    }
  }
}

static int
append_arc (pdf_path  *p,
	    pdf_coord *cp,
	    double c_x , double c_y,
	    double r_x , double r_y,
	    double ang0, double ang1, int dir)
{
  pdf_tmatrix M;
  pdf_coord   p0, p1, p2;
  int    i, n_steps;
  double delta, v[6];

  if (!p || !cp)
    return 1;

  if (dir >= 0) {
    while (ang1 < ang0)
      ang0 -= 360;
  } else {
    while (ang1 > ang0)
      ang1 -= 360;
  }

  ang0 *= 3.1415926/180;
  ang1 *= 3.1415926/180;

#define MAX_ANG_STEP 1.0
  if (dir >= 0) {
    n_steps = (int) ((ang1 - ang0)/MAX_ANG_STEP) + 1;
  } else {
    n_steps = (int) ((ang0 - ang1)/MAX_ANG_STEP) + 1;
  }
  delta = (ang1 - ang0) / n_steps;

  M.a =  cos(delta); M.b = sin(delta);
  M.c = -sin(delta); M.d = cos(delta);
  M.e = 0.0;
  M.f = 0.0;

  p0.x = cos(ang0); p0.y = sin(ang0);

  p1.x = p0.x - delta / 3.0 * p0.y;
  p1.y = p0.y + delta / 3.0 * p0.x;
  p2.x = p0.x + delta / 3.0 * p0.y;
  p2.y = p0.y - delta / 3.0 * p0.x;

  v[0] = c_x + r_x * p0.x;
  v[1] = c_y + r_y * p0.y;
  if (p->num_paths == 0) {
    pdf_draw_append_path(p, cp, PDF_PATH_MOVETO, v);
  } else if (cp->x != v[0] ||
	     cp->y != v[1]) {
    pdf_draw_append_path(p, cp, PDF_PATH_LINETO, v);
  }

  pdf_draw_transform_coord(&p2, &M);
  pdf_draw_transform_coord(&p0, &M);
  for (i = 0; i < n_steps; i++) {
    v[0] = c_x + r_x * p1.x;
    v[1] = c_y + r_y * p1.y;
    v[2] = c_x + r_x * p2.x;
    v[3] = c_y + r_y * p2.y;
    v[4] = c_x + r_x * p0.x;
    v[5] = c_y + r_y * p0.y;
    pdf_draw_append_path(p, cp, PDF_PATH_CURVETO, v);
    pdf_draw_transform_coord(&p0, &M);
    pdf_draw_transform_coord(&p1, &M);
    pdf_draw_transform_coord(&p2, &M);
  }

  return 0;
}


int
pdf_draw_append_arc (pdf_path  *p,
		     pdf_coord *cp,
		     double c_x , double c_y,
		     double r_x , double r_y,
		     double ang0, double ang1)
{
  return (append_arc(p, cp, c_x, c_y, r_x, r_y, ang0, ang1,  1));
}

int
pdf_draw_append_arcn (pdf_path  *p,
		      pdf_coord *cp,
		      double c_x , double c_y,
		      double r_x , double r_y,
		      double ang0, double ang1)
{
  return (append_arc(p, cp, c_x, c_y, r_x, r_y, ang0, ang1, -1));
}

int
pdf_draw_append_bspline (pdf_path  *p, pdf_coord *cp,
			 double x0, double y0, double x1, double y1, double x2, double y2)
{
  double v[6];

  if (!p || !cp)
    return 0;

  v[0] = x0 + 2.0 * (x1 - x0) / 3.0;
  v[1] = y0 + 2.0 * (y1 - y0) / 3.0;
  v[2] = x1 + (x2 - x1) / 3.0;
  v[3] = y1 + (y2 - y1) / 3.0;
  v[4] = x2;
  v[5] = y2;

  pdf_draw_append_path(p, cp, PDF_PATH_CURVETO, v);

  return 0;
}

