/*  $Header: /home/cvsroot/dvipdfmx/src/Attic/tpic.c,v 1.13 2004/08/07 09:52:34 hirata Exp $

    This is dvipdfmx, an eXtended version of dvipdfm by Mark A. Wicks.

    Copyright (C) 2002 by Jin-Hwan Cho and Shunsaku Hirata,
    the dvipdfmx project team <dvipdfmx@project.ktug.or.kr>
    
    Copyright (C) 1998, 1999 by Mark A. Wicks <mwicks@kettering.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>

#include "system.h"
#include "mem.h"
#include "mfileio.h"

#include "pdfresource.h"
#include "pdfparse.h"
#include "pdfdoc.h"
#include "pdfdev.h"
#include "dvi.h"

#include "pdfdraw.h"
#include "tpic.h"

/*
 * Following "constant" converts milli-inches to
 * device (in this case PDF stream) coordinates.
 */

#define MI2DEV (0.072/pdf_dev_scale())

static struct
{
  double     pen_size;
  int        fill_shape;
  double     fill_color;
  pdf_coord *path;
  long       path_length;
  long       max_path_length;
} tp_gstate = {
  1.0, 0, 0.0, NULL, 0, 0
};

#define TPIC_DEFAULT_FILL_MODE 0

static double default_fill_color = 0.5;
static int    tp_fill_mode = TPIC_DEFAULT_FILL_MODE;

void
tpic_set_fill_mode (int mode)
{
  int pdf_ver;

  pdf_ver = pdf_get_version();
  if (mode == 1) {
    if (pdf_ver >= 4)
      tp_fill_mode = 1;
    else {
      WARN("TPIC: tpic shading support requires PDF version 1.4.");
      tp_fill_mode = 0;
    }
  } else {
    tp_fill_mode = 0;
  }
}

static void
tpic_clear_state (void) 
{
  if (tp_gstate.path) {
    RELEASE(tp_gstate.path);
    tp_gstate.path = NULL;
  }
  tp_gstate.path_length     = 0;
  tp_gstate.max_path_length = 0;
  tp_gstate.fill_shape = 0;
  tp_gstate.fill_color = 0.0;
}

static void
set_pen_size (char **buffer, char *end)
{
  char *number;

  skip_white (buffer, end);
  number = parse_number(buffer, end);
  if (!number)
    WARN("TPIC: Invalid pen size.");
  else {
    tp_gstate.pen_size = atof (number) * MI2DEV;
    RELEASE(number);
  }
}

static void
set_fill_color (char **buffer, char *end)
{
  char *number;

  tp_gstate.fill_shape = 1;
  tp_gstate.fill_color = default_fill_color;

  skip_white(buffer, end);
  number = parse_number(buffer, end);
  if (number) {
    tp_gstate.fill_color = 1.0 - atof(number);
    RELEASE(number);
  }

  if (tp_gstate.fill_color > 1.0)
    tp_gstate.fill_color = 1.0;
  if (tp_gstate.fill_color < 0.0)
    tp_gstate.fill_color = 0.0;
}

static void
add_point (char **buffer, char *end) 
{
  char *x= NULL, *y= NULL;

  skip_white(buffer, end);
  if (*buffer < end)
    x = parse_number(buffer, end);
  skip_white(buffer, end);
  if (*buffer < end)
    y = parse_number(buffer, end);
  if ((x) && (y)) {
    if (tp_gstate.path_length >= tp_gstate.max_path_length) {
      tp_gstate.max_path_length += 256;
      tp_gstate.path = RENEW(tp_gstate.path, tp_gstate.max_path_length, pdf_coord);
    }
    tp_gstate.path[tp_gstate.path_length].x = atof(x)*MI2DEV;
    tp_gstate.path[tp_gstate.path_length].y = atof(y)*MI2DEV;
    tp_gstate.path_length += 1;
  } else {
    WARN("TPIC: Missing coordinate for \"pa\".");
  }
  if (x) RELEASE(x);
  if (y) RELEASE(y);

  return;
}


/*
 * We can include line-width, dash pattern and other graphic state parameters
 * in ExtGState dictionary. But doing that is not efficient.
 * "g" is gray level.
 */
static void
set_shading (double g)
{
  pdf_obj *gs_obj, *gs_ref;
  char res_name[6]; /* GsXX */
  char wbuf[32];
  int  alpha, len = 0;

  /*
   * The semantics of a fill_color of 0.0 or 0.5 will be to
   * use current painting color known to dvipdfm.
   */
  if (tp_fill_mode != 1) {
    if (g != 0.0) {
      len = sprintf(wbuf, " %.2f g", g);
    }
  } else {
    if (g == 1.0)
      pdf_doc_add_to_page(" 1 g", 4);
    else {
      alpha = ROUND(255.0*(1.0-g), 1.0);
      sprintf(res_name, "GsT%02X", alpha);

      gs_ref = PDF_findresource(res_name, PDF_RES_SUBTYPE_GSTATE);
      if (!gs_ref) {
	gs_obj = pdf_new_dict();
	/* Type, BM, ... optional (omitted) */
	pdf_add_dict(gs_obj, pdf_new_name("ca"), pdf_new_number(alpha/255.0));
	gs_ref = PDF_defineresource(res_name, gs_obj, PDF_RES_SUBTYPE_GSTATE);
      }
      pdf_doc_add_to_page_resources("ExtGState",
				    res_name, pdf_link_obj(gs_ref));
      len = sprintf(wbuf, " 0 g /%s gs", res_name);
    }
  }
  pdf_doc_add_to_page(wbuf, len);
}


static void
show_path (pdf_path *p, int hidden) 
{
  if (tp_gstate.fill_shape) {
    set_shading(tp_gstate.fill_color);
  }

  if (!hidden) {
    if (tp_gstate.fill_shape)
      pdf_draw_path(p, "b");
    else
      pdf_draw_path(p, "S");
  } else {
    /*
     * Acrobat claims 'Q' as illegal operation when there are unfinished
     * path (a path without path-painting operator applied)?
     */
    if (tp_gstate.fill_shape)
      pdf_draw_path(p, "f");
    else
      pdf_draw_path(p, "n");
  }
  tp_gstate.fill_color = 0.0;
  tp_gstate.fill_shape = 0;
}

#define CLOSED_PATH() (tp_gstate.path[0].x == tp_gstate.path[tp_gstate.path_length-1].x && \
		       tp_gstate.path[0].y == tp_gstate.path[tp_gstate.path_length-1].y)

static void
flush_path (double x_user, double y_user, int hidden, double dash_dot)
{
  pdf_path *p;
  pdf_coord cp;
  double    v[2];
  int       i;

  /* Make pen_size == 0 equivalent to hidden */
  if (tp_gstate.pen_size == 0.0)
    hidden = 1;

  if (tp_gstate.path_length <= 1) {
    WARN("TPIC: Not enough points for \"fp\"!");
  } else if (!hidden || tp_gstate.fill_shape) {
    p = pdf_draw_new_path();

    pdf_doc_add_to_page(" q", 2);
    pdf_draw_setlinecap (1);
    pdf_draw_setlinejoin(1);
    if (tp_gstate.pen_size != 0.0)
      pdf_draw_setlinewidth(tp_gstate.pen_size);
    if (dash_dot > 0.0) {
      v[0] = dash_dot*72.0;
      pdf_draw_setdash(1, v, 0);
    } else if (dash_dot < 0.0) {
      v[0] =  tp_gstate.pen_size;
      v[1] = -dash_dot*72.0;
      pdf_draw_setdash(2, v, 0);
    }

    v[0] = cp.x = x_user + tp_gstate.path[0].x;
    v[1] = cp.y = y_user - tp_gstate.path[0].y;
    pdf_draw_append_path(p, &cp, PDF_PATH_MOVETO, v);

    for (i = 0;
	 i < tp_gstate.path_length; i++) {
      v[0] = x_user + tp_gstate.path[i].x;
      v[1] = y_user - tp_gstate.path[i].y;
      pdf_draw_append_path(p, &cp, PDF_PATH_LINETO, v);
    } 
    /* Shading is applied only to closed path. */
    if (!CLOSED_PATH())
      tp_gstate.fill_shape = 0;
    show_path(p, hidden);
    pdf_doc_add_to_page(" Q", 2);

    pdf_draw_release_path(p);
  }
  tpic_clear_state();

  return;
}


/*
 * Accroding to
 * "Tpic: Pic for TEX", Tim Morgan, Original by Brian Kernighan, p.20:
 * 
 *  A spline is a smooth curve guided by a set of straight lines just
 *  like the line above. It begins at the same place, ends at the same
 *  place, and in between is tangent to the mid-point of each guiding
 *  line. The syntax for a spline is identical to a (path) line except
 *  for using spline instead of line.
 *
 * Spline is not a curve drawn by spline-fitting points p0, p1, ..., pn,
 * defined by tpic special "pa" command. Instead, a path defined by set
 * of points p0, p1, ... is guiding line mentioned above.
 *
 * Dvipsk draws them as a straight line from p0 to q1 = (p0 + p1)/2,
 * followed by a quadratic B-spline curve with starting point q1, (off-
 * curve) control point p1, end point q2 = (p1 + p2)/2, ..., and a
 * straight line from qn to pn.
 */

static void
spline_path (double x_user, double y_user, double dash_dot)
{
  pdf_path *p;
  pdf_coord cp;
  int       i;
  double    v[6];

  /* Spline is meaningless for path length of less than 3 */
  if (tp_gstate.path_length <= 2)
    WARN("TPIC: Not enough points for \"sp\"!");
  else {
    p = pdf_draw_new_path();

    pdf_doc_add_to_page(" q", 2);
    pdf_draw_setmiterlimit(1.4);
    if (tp_gstate.pen_size != 0.0)
      pdf_draw_setlinewidth (tp_gstate.pen_size);
    if (dash_dot > 0.0) {
      v[0] = dash_dot*72.0;
      pdf_draw_setdash(1, v, 0);
    } else if (dash_dot < 0.0) {
      v[0] =  tp_gstate.pen_size;
      v[1] = -dash_dot*72.0;
      pdf_draw_setdash(2, v, 0);
      pdf_draw_setlinecap(1);
    }

    v[0] = cp.x = x_user + tp_gstate.path[0].x;
    v[1] = cp.y = y_user - tp_gstate.path[0].y;
    pdf_draw_append_path(p, &cp, PDF_PATH_MOVETO, v);

    v[0] = x_user + 0.5*(tp_gstate.path[0].x + tp_gstate.path[1].x);
    v[1] = y_user - 0.5*(tp_gstate.path[0].y + tp_gstate.path[1].y);
    pdf_draw_append_path(p, &cp, PDF_PATH_LINETO, v);

    for (i = 1; i < tp_gstate.path_length - 1; i++) {
      /* B-spline control points */

      v[0] = x_user + 0.5 * (tp_gstate.path[i-1].x + tp_gstate.path[i].x);
      v[1] = y_user - 0.5 * (tp_gstate.path[i-1].y + tp_gstate.path[i].y);
      v[2] = x_user + tp_gstate.path[i].x;
      v[3] = y_user - tp_gstate.path[i].y;
      v[4] = x_user + 0.5 * (tp_gstate.path[i].x + tp_gstate.path[i+1].x);
      v[5] = y_user - 0.5 * (tp_gstate.path[i].y + tp_gstate.path[i+1].y);

      pdf_draw_append_bspline(p, &cp, v[0], v[1], v[2], v[3], v[4], v[5]);
    } 

    v[0] = x_user + tp_gstate.path[i].x;
    v[1] = y_user - tp_gstate.path[i].y;
    pdf_draw_append_path(p, &cp, PDF_PATH_LINETO, v);

    if (!CLOSED_PATH())
      tp_gstate.fill_shape = 0;

    show_path(p, 0);

    pdf_doc_add_to_page(" Q", 2);

    pdf_draw_release_path(p);
  }

  tpic_clear_state();

  return;
}

static void
arc (char **buffer, char *end, double x_user, double y_user, int hidden) 
{
  pdf_path *p;
  pdf_coord cp;
  char  *xcs, *ycs, *xrs, *yrs, *sas, *eas;
  double xc, yc, xr, yr, sa, ea;

  if (tp_gstate.pen_size == 0.0)
    hidden = 1;

  xcs = ycs = xrs = yrs = sas = eas = NULL;
  if (!(xcs = parse_number(buffer, end)) ||
      !(ycs = parse_number(buffer, end)) ||
      !(xrs = parse_number(buffer, end)) ||
      !(yrs = parse_number(buffer, end)) ||
      !(sas = parse_number(buffer, end)) ||
      !(eas = parse_number(buffer, end))) {
    WARN("TPIC: Not enough parameters for ar/ir command.");
  } else if (!hidden || tp_gstate.fill_shape) {
    xc = atof(xcs) * MI2DEV; yc = atof(ycs) * MI2DEV;
    xr = atof(xrs) * MI2DEV; yr = atof(yrs) * MI2DEV;
    sa = atof(sas) * 180.0 / 3.1415926;
    ea = atof(eas) * 180.0 / 3.1415926;

    p = pdf_draw_new_path();
    pdf_doc_add_to_page(" q", 2);
    if (tp_gstate.pen_size != 0.0) {
      pdf_draw_setlinewidth(tp_gstate.pen_size);
    }
    pdf_draw_setlinecap(1);

    pdf_draw_append_arcn(p, &cp, x_user + xc, y_user - yc, xr, yr, -sa, -ea);

    if ((int) (fabs(ea-sa)+0.5) < 360)
      tp_gstate.fill_shape = 0;

    show_path(p, hidden);

    pdf_doc_add_to_page(" Q", 2);

    pdf_draw_release_path(p);
  }
  if (xcs) RELEASE(xcs);
  if (ycs) RELEASE(ycs);
  if (xrs) RELEASE(xrs);
  if (yrs) RELEASE(yrs);
  if (sas) RELEASE(sas);
  if (eas) RELEASE(eas);

  tpic_clear_state();

  return;
}

#define TPIC_PN 1
#define TPIC_PA 2
#define TPIC_FP 3
#define TPIC_IP 4
#define TPIC_DA 5
#define TPIC_DT 6
#define TPIC_SP 7
#define TPIC_AR 8
#define TPIC_IA 9
#define TPIC_SH 10
#define TPIC_WH 11
#define TPIC_BK 12
#define TPIC_TX 13

#define TPIC_UNKNOWN -1

static struct {
  const char *s;
  int tpic_command;
} tpic_specials[] = {
  {"pn", TPIC_PN},
  {"pa", TPIC_PA},
  {"fp", TPIC_FP},
  {"ip", TPIC_IP},
  {"da", TPIC_DA},
  {"dt", TPIC_DT},
  {"sp", TPIC_SP},
  {"ar", TPIC_AR},
  {"ia", TPIC_IA},
  {"sh", TPIC_SH},
  {"wh", TPIC_WH},
  {"bk", TPIC_BK},
  {"tx", TPIC_TX},
  {NULL, -1}
};

int
tpic_parse_special (char *buffer, UNSIGNED_QUAD size,
		    double x_user, double y_user)
{
  int   result;
  int   tpic_command;
  char *end = buffer + size;
  char *token;

  result = 0;
  tpic_command = TPIC_UNKNOWN;
  skip_white(&buffer, end);
  token = parse_ident(&buffer, end);
  if (token) {
    int i;
    for (i = 0; tpic_specials[i].s != NULL; i++) {
      if (!strcmp(tpic_specials[i].s, token)) {
	tpic_command = tpic_specials[i].tpic_command;
	break;
      }
    }
    RELEASE(token);
  }
  if (!token || tpic_command == TPIC_UNKNOWN)
    return 0;

  result = 1;
  skip_white(&buffer, end);
  switch (tpic_command) {
  case TPIC_PN:
    set_pen_size(&buffer, end);
    break;
  case TPIC_PA:
    add_point(&buffer, end);
    break;
  case TPIC_FP:
    flush_path(x_user, y_user, 0, 0.0);
    break;
  case TPIC_IP: 
    flush_path(x_user, y_user, 1, 0.0);
    break;
  case TPIC_DA:
    token = parse_number(&buffer, end);
    if (token) {
      flush_path(x_user, y_user, 0, atof(token));
      RELEASE(token);
    }
    break;
  case TPIC_DT:
    token = parse_number(&buffer, end);
    if (token) {
      flush_path(x_user, y_user, 0, -atof(token));
      RELEASE(token);
    }
    break;
  case TPIC_SP:
    token = parse_number(&buffer, end);
    if (token) {
      spline_path(x_user, y_user, atof(token));
      RELEASE(token);
    } else {
      spline_path(x_user, y_user, 0.0);
    }
    break;
  case TPIC_AR:
    arc(&buffer, end, x_user, y_user, 0);
    break;
  case TPIC_IA:
    arc(&buffer, end, x_user, y_user, 1);
    break;
  case TPIC_SH:
    set_fill_color(&buffer, end);
    break;
  case TPIC_WH: 
    tp_gstate.fill_shape = 1;
    tp_gstate.fill_color = 1.0;
    break;
  case TPIC_BK:
    tp_gstate.fill_shape = 1;
    tp_gstate.fill_color = 0.0;
    break;
  case TPIC_TX: 
    tp_gstate.fill_shape = 1;
    {
      long num = 0, den = 0;
      while (buffer++ < end) {
	switch (*(buffer++)) {
	case '0':
	  num += 0;
	case '1': case '2': case '4': case '8':
	  num += 1;
	  break;
	case '3': case '5': case '6': case '9':
	case 'a': case 'A': case 'c': case 'C':
	  num += 2;
	  break;
	case '7': case 'b': case 'B': case 'd':
	case 'D':
	  num += 3;
	  break;
	case 'f': case 'F':
	  num += 4;
	  break;
	default:
	  break;
	}
	den += 16;
      }
      if (den != 0) {
	default_fill_color = 1.0 - (float) (num)/(den);
      }
      else {
	default_fill_color = 0.5;
      }
    }
    break;
  default:
    ERROR("TPIC: Fix me, I'm broke. This should never happen");
  }

  return result;
}
