/*  $Header: /home/cvsroot/dvipdfmx/src/type1.c,v 1.32 2004/08/15 15:37:08 hirata Exp $

    This is dvipdfmx, an eXtended version of dvipdfm by Mark A. Wicks.

    Copyright (C) 2002 by Jin-Hwan Cho and Shunsaku Hirata,
    the dvipdfmx project team <dvipdfmx@project.ktug.or.kr>

    Copyright (C) 1998, 1999 by Mark A. Wicks <mwicks@kettering.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <math.h>

#include "system.h"
#include "mfileio.h"
#include "mem.h"
#include "error.h"
#include "numbers.h"

#include "pdfobj.h"
#include "pdffont.h"

#include "encodings.h"
#include "unicode.h"

#include "dpxutil.h"

#include "t1crypt.h"
#include "pst_obj.h"
#include "pst.h"

#include "cff_limits.h"
#include "cff_types.h"
#include "cff_dict.h"
#include "cff.h"

#include "t1_load.h"
#include "t1_char.h"

#include "type1.h"

#define TYPE1_NAME_LEN_MAX   127

#define FONT_FLAG_FIXEDPITCH (1 << 0)  /* Fixed-width font */
#define FONT_FLAG_SERIF      (1 << 1)  /* Serif font */
#define FONT_FLAG_SYMBOLIC   (1 << 2)  /* Symbolic font */
#define FONT_FLAG_SCRIPT     (1 << 3)  /* Script font */
#define FONT_FLAG_STANDARD   (1 << 5)  /* Adobe Standard Character Set */
#define FONT_FLAG_ITALIC     (1 << 6)  /* Italic */
#define FONT_FLAG_ALLCAP     (1 << 16) /* All-cap font */
#define FONT_FLAG_SMALLCAP   (1 << 17) /* Small-cap font */
#define FONT_FLAG_FORCEBOLD  (1 << 18) /* Force bold at small text sizes */

#define TYPE1FONT_FLAG_BASEFONT (1 << 0)

static int
is_basefont (const char *name)
{
  static const char *basefonts[] = {
    "Courier",                  "Courier-Bold",          "Courier-Oblique",
    "Courier-BoldOblique",      "Helvetica",             "Helvetica-Bold",
    "Helvetica-Oblique",        "Helvetica-BoldOblique", "Symbol",
    "Times-Roman",              "Times-Bold",            "Times-Italic",
    "Times-BoldItalic",         "ZapfDingbats"
  };
  int i;

  for (i = 0; i < 14; i++) {
    if (!strcmp(name, basefonts[i]))
      return 1;
  }

  return 0;
}

static int
Type1Font_is_basefont (pdf_font *font)
{
  ASSERT(font);

  return (font->flags & TYPE1FONT_FLAG_BASEFONT);
}

int
Type1Font_match (pdf_font *font, const char *name,
		 int encoding_id, double font_scale, double design_size, int embedding)
{
  ASSERT(font);

  if (name && font->ident && !strcmp(font->ident, name) &&
      encoding_id == font->encoding_id &&
      embedding == font->embedding)
    return 1;

  return 0;
}

int
Type1Font_open (pdf_font *font, const char *name,
		int encoding_id, double font_scale, double design_size, int embedding)
{
  FILE *fp;
  char *fullname = NULL;
  char  fontname[PDF_NAME_LEN_MAX+8]; /* 7 bytes for pseudo unique tag */

  ASSERT(font);

  if (is_basefont(name)) {
    font->ident = NEW(strlen(name)+1, char);
    strcpy(font->ident, name);
    if (strlen(name) > PDF_NAME_LEN_MAX)
      ERROR("Unexpected error...");
    strcpy(font->fontname, name);
    font->filename    = NULL;
    font->encoding_id = encoding_id;
    font->embedding   = 0;
    font->fontdict    = pdf_new_dict();
    font->descriptor  = NULL;

    font->flags |= TYPE1FONT_FLAG_BASEFONT;

    pdf_add_dict(font->fontdict,
		 pdf_new_name("Type"), pdf_new_name("Font"));
    pdf_add_dict(font->fontdict,
		 pdf_new_name("Subtype"), pdf_new_name("Type1"));
    pdf_add_dict(font->fontdict,
		 pdf_new_name("BaseFont"), pdf_new_name(font->fontname));
    return 0;
  }

  fullname = kpse_find_file(name, kpse_type1_format, 0);
  if (!fullname)
    return -1;
  fp = MFOPEN(fullname, FOPEN_RBIN_MODE);
  if (!fp) {
    WARN("Failed to open Type 1 font \"%s\".", fullname);
    return -1;
  }

  memset(fontname, 0, PDF_NAME_LEN_MAX+1);
  if (!is_pfb(fp) || t1_get_fontname(fp, fontname) < 0) {
    MFCLOSE(fp);
    return -1;
  }
  MFCLOSE(fp);

  font->ident = NEW(strlen(name)+1, char);
  strcpy(font->ident, name);
  strcpy(font->fontname, fontname);
  font->filename = NEW(strlen(fullname)+1, char);
  strcpy(font->filename, fullname);
  font->encoding_id = encoding_id;
  font->embedding   = embedding;
  font->fontdict    = pdf_new_dict();
  font->descriptor  = pdf_new_dict();

  pdf_add_dict(font->fontdict,
	       pdf_new_name("Type"), pdf_new_name("Font"));
  pdf_add_dict(font->fontdict,
	       pdf_new_name("Subtype"), pdf_new_name("Type1"));
  pdf_add_dict(font->descriptor,
	       pdf_new_name("Type"), pdf_new_name("FontDescriptor"));

  if (font->embedding) {
    pdf_font_make_tag(font->uniqueID);
    sprintf(fontname, "%6s+%s", font->uniqueID, font->fontname);
  } else {
    strcpy(fontname, font->fontname);
  }

  pdf_add_dict(font->fontdict,
	       pdf_new_name("BaseFont"), pdf_new_name(fontname));
  pdf_add_dict(font->descriptor,
	       pdf_new_name("FontName"), pdf_new_name(fontname));

  return 0;
}

static void
get_font_attr (pdf_font *font, cff_font *cff)
{
  double capheight, ascent, descent;
  double italicangle, stemv;
  double defaultwidth, nominalwidth;
  long   flags = 0;
  long   gid;
  int    i;
  static const char *L_c[] = {
    "H", "P", "Pi", "Rho", NULL
  };
  static const char *L_d[] = {
    "p", "q", "mu", "eta", NULL
  };
  static const char *L_a[] = {
    "b", "h", "lambda", NULL
  };
  t1_ginfo gm;

  defaultwidth = 500.0;
  nominalwidth = 0.0;

  /*
   * CapHeight, Ascent, and Descent is meaningfull only for Latin/Greek/Cyrillic.
   * The BlueValues and OtherBlues also have those information.
   */
  if (cff_dict_known(cff->topdict, "FontBBox")) {
    /* Default values */
    capheight = ascent = cff_dict_get(cff->topdict, "FontBBox", 3);
    descent = cff_dict_get(cff->topdict, "FontBBox", 1);
  } else {
    capheight =  680.0;
    ascent    =  690.0;
    descent   = -190.0;
  }
  if (cff_dict_known(cff->private[0], "StdVW")) {
    stemv = cff_dict_get(cff->private[0], "StdVW", 0);
  } else {
    /*
     * We may use the following values for StemV:
     *  Thin - ExtraLight: <= 50
     *  Light: 71
     *  Regular(Normal): 88
     *  Medium: 109
     *  SemiBold(DemiBold): 135
     *  Bold - Heavy: >= 166
     */
    stemv = 88.0;
  }
  if (cff_dict_known(cff->topdict, "ItalicAngle")) {
    italicangle = cff_dict_get(cff->topdict, "ItalicAngle", 0);
    if (italicangle != 0.0)
      flags |= FONT_FLAG_ITALIC;
  } else {
    italicangle = 0.0;
  }

  /*
   * Use "space", "H", "p", and "b" for various values.
   * Those characters should not "seac". (no accent)
   */
  gid = cff_glyph_lookup(cff, "space");
  if (gid >= 0 && gid < cff->cstrings->count) {
    t1char_get_metrics(cff->cstrings->data + cff->cstrings->offset[gid] - 1,
		       cff->cstrings->offset[gid+1] - cff->cstrings->offset[gid+1],
		       cff->subrs[0], &gm);
    defaultwidth = gm.wx;
  }

  for (i = 0; L_c[i] != NULL; i++) {
    gid = cff_glyph_lookup(cff, L_c[i]);
    if (gid >= 0 && gid < cff->cstrings->count) {
      t1char_get_metrics(cff->cstrings->data + cff->cstrings->offset[gid] - 1,
		      cff->cstrings->offset[gid+1] - cff->cstrings->offset[gid+1],
		      cff->subrs[0], &gm);
      capheight = gm.bbox.ury;
      break;
    }
  }

  for (i = 0; L_d[i] != NULL; i++) {
    gid = cff_glyph_lookup(cff, L_d[i]);
    if (gid >= 0 && gid < cff->cstrings->count) {
      t1char_get_metrics(cff->cstrings->data + cff->cstrings->offset[gid] - 1,
			 cff->cstrings->offset[gid+1] - cff->cstrings->offset[gid+1],
			 cff->subrs[0], &gm);
      descent = gm.bbox.lly;
      break;
    }
  }

  for (i = 0; L_a[i] != NULL; i++) {
    gid = cff_glyph_lookup(cff, L_a[i]);
    if (gid >= 0 && gid < cff->cstrings->count) {
      t1char_get_metrics(cff->cstrings->data + cff->cstrings->offset[gid] - 1,
			 cff->cstrings->offset[gid+1] - cff->cstrings->offset[gid+1],
			 cff->subrs[0], &gm);
      ascent = gm.bbox.ury;
      break;
    }
  }

  if (defaultwidth != 0.0) {
    cff_dict_add(cff->private[0], "defaultWidthX", 1);
    cff_dict_set(cff->private[0], "defaultWidthX", 0, defaultwidth);
  }
  if (nominalwidth != 0.0) {
    cff_dict_add(cff->private[0], "nominalWidthX", 1);
    cff_dict_set(cff->private[0], "nominalWidthX", 0, nominalwidth);
  }
  if (cff_dict_known(cff->private[0], "ForceBold") &&
      cff_dict_get(cff->private[0], "ForceBold", 0)) {
    flags |= FONT_FLAG_FORCEBOLD;
  }
  if (cff_dict_known(cff->private[0], "IsFixedPitch") &&
      cff_dict_get(cff->private[0], "IsFixedPitch", 0)) {
    flags |= FONT_FLAG_FIXEDPITCH;
  }
  if (font->fontname &&
      !strstr(font->fontname, "Sans")) {
    flags |= FONT_FLAG_SERIF;
  }
  flags |= FONT_FLAG_SYMBOLIC;

  pdf_add_dict(font->descriptor,
	       pdf_new_name("CapHeight"), pdf_new_number(capheight));
  pdf_add_dict(font->descriptor,
	       pdf_new_name("Ascent"), pdf_new_number(ascent));
  pdf_add_dict(font->descriptor,
	       pdf_new_name("Descent"), pdf_new_number(descent));
  pdf_add_dict(font->descriptor,
	       pdf_new_name("ItalicAngle"), pdf_new_number(italicangle));
  pdf_add_dict(font->descriptor,
	       pdf_new_name("StemV"), pdf_new_number(stemv));
  pdf_add_dict(font->descriptor,
	       pdf_new_name("Flags"), pdf_new_number(flags));
}

static void
add_metrics (pdf_font *font, cff_font *cff, double *widths, long num_glyphs)
{
  pdf_obj *tmp;
  int    code, firstchar, lastchar;
  double val;
  card16 gid;
  int    i;

  /*
   * The original FontBBox of the font is preserved, instead
   * of replacing it with tight bounding box calculated from
   * charstrings, to prevent Acrobat 4 from greeking text as
   * much as possible.
   */
  if (!cff_dict_known(cff->topdict, "FontBBox")) {
    ERROR("No FontBBox?");
  }
  tmp = pdf_new_array();
  for (i = 0; i < 4; i++) {
    val = cff_dict_get(cff->topdict, "FontBBox", i);
    pdf_add_array(tmp, pdf_new_number(ROUND(val, 1.0)));
  }
  pdf_add_dict(font->descriptor, pdf_new_name("FontBBox"), tmp);

  tmp = pdf_new_array();
  if (num_glyphs <= 1) {
    /* This should be error. */
    firstchar = lastchar = 0;
    pdf_add_array(tmp, pdf_new_number(0.0));
  } else {
    firstchar = 255; lastchar = 0;
    for (code = 0; code < 256; code++) {
      if (font->used_chars[code]) {
	if (code < firstchar)
	  firstchar = code;
	if (code > lastchar)
	  lastchar  = code;
      }
    }
    for (code = firstchar; code <= lastchar; code++) {
      if (font->used_chars[code]) {
	gid = cff_encoding_lookup(cff, code);
	pdf_add_array(tmp, pdf_new_number(ROUND(widths[gid], 1.0)));
      } else {
	pdf_add_array(tmp, pdf_new_number(0.0));
      }
    }
  }
  pdf_add_dict(font->fontdict,
	       pdf_new_name("FirstChar"), pdf_new_number(firstchar));
  pdf_add_dict(font->fontdict,
	       pdf_new_name("LastChar"),  pdf_new_number(lastchar));
  pdf_add_dict(font->fontdict, pdf_new_name("Widths"), pdf_ref_obj(tmp));
  pdf_release_obj(tmp);
}

static long
write_fontfile (pdf_font *font, cff_font *cff, long num_glyphs)
{
  cff_index *topdict;
  long private_size, destlen, charstring_len;
  long topdict_offset, offset;
  card8 *dest;

  {
#define WBUF_SIZE 1024
    card8 wbuf[WBUF_SIZE];

    topdict = cff_new_index(1);
    /*
     * Force existence of Encoding.
     */
    if (!cff_dict_known(cff->topdict, "CharStrings"))
      cff_dict_add(cff->topdict, "CharStrings", 1);
    if (!cff_dict_known(cff->topdict, "charset"))
      cff_dict_add(cff->topdict, "charset", 1);
    if (!cff_dict_known(cff->topdict, "Encoding"))
      cff_dict_add(cff->topdict, "Encoding", 1);
    private_size = cff_dict_pack((cff->private)[0], wbuf, WBUF_SIZE);
    if (private_size > 0 && !cff_dict_known(cff->topdict, "Private"))
      cff_dict_add(cff->topdict, "Private", 2);
    (topdict->offset)[1] = cff_dict_pack(cff->topdict, wbuf, WBUF_SIZE) + 1;
  }
  /*
   * Estimate total size of fontfile.
   */
  charstring_len = cff_index_size(cff->cstrings);
  {
    destlen = 4; /* header size */
    destlen += cff_index_size(cff->name);
    destlen += cff_index_size(topdict);
    destlen += cff_index_size(cff->string);
    destlen += cff_index_size(cff->gsubr);
    /*
     * We are using format 1 for Encoding and format 0 for charset.
     * TODO: Should implement cff_xxx_size().
     */
    destlen += 2 + (cff->encoding->num_entries)*2 + 1 + (cff->encoding->num_supps)*3;
    destlen += 1 + (cff->charsets->num_entries)*2;
    destlen += charstring_len;
    destlen += private_size;
  }
  /*
   * Now we create FontFile data.
   */
  {
    dest = NEW(destlen, card8);
    /*
     * Data Layout order as described in CFF spec., sec 2 "Data Layout".
     */
    offset = 0;
    /* Header */
    offset += cff_put_header(cff, dest + offset, destlen - offset);
    /* Name */
    offset += cff_pack_index(cff->name, dest + offset, destlen - offset);
    /* Top DICT */
    topdict_offset = offset;
    offset += cff_index_size(topdict);
    /* Strings */
    offset += cff_pack_index(cff->string, dest + offset, destlen - offset);
    /* Global Subrs */
    offset += cff_pack_index(cff->gsubr, dest + offset, destlen - offset);
    /* Encoding */
    cff_dict_set(cff->topdict, "Encoding", 0, offset);
    offset += cff_pack_encoding(cff, dest + offset, destlen - offset);
    /* charset */
    cff_dict_set(cff->topdict, "charset", 0, offset);
    offset += cff_pack_charsets(cff, dest + offset, destlen - offset);
    /* CharStrings */
    cff_dict_set(cff->topdict, "CharStrings", 0, offset);
    offset += cff_pack_index(cff->cstrings, dest + offset, charstring_len);
    /* Private */
    if ((cff->private)[0] && private_size > 0) {
      private_size = cff_dict_pack((cff->private)[0], dest + offset, private_size);
      cff_dict_set(cff->topdict, "Private", 1, offset);
      cff_dict_set(cff->topdict, "Private", 0, private_size);
    }
    offset += private_size;

    /* Finally Top DICT */
    topdict->data = NEW(topdict->offset[1] - 1, card8);
    cff_dict_pack(cff->topdict, topdict->data, topdict->offset[1] - 1);
    cff_pack_index(topdict, dest + topdict_offset, cff_index_size(topdict));
    cff_release_index(topdict);

    /* Copyright and Trademark Notice ommited. */

    {
      pdf_obj *fontfile, *stream_dict;

      fontfile    = pdf_new_stream(STREAM_COMPRESS);
      stream_dict = pdf_stream_dict(fontfile);
      pdf_add_dict(font->descriptor,
		   pdf_new_name("FontFile3"),
		   pdf_ref_obj (fontfile));
      pdf_add_dict(stream_dict,
		   pdf_new_name("Subtype"),
		   pdf_new_name("Type1C"));
      pdf_add_stream (fontfile, (char *) dest, offset);
      pdf_release_obj(fontfile);
      RELEASE(dest);
    }
  }

  return offset;
}


void
Type1Font_dofont (pdf_font *font)
{
  cff_font     *cff;
  cff_charsets *charset;
  char  **enc_vec;
  double  defaultwidth, nominalwidth;
  double *widths;
  card16 *GIDMap, num_glyphs = 0;
  FILE *fp;
  long  offset;
  int   cc, verbose;

  ASSERT(font);

  if (!font->indirect)
    return;

  verbose = pdf_font_get_verbose();

  if (Type1Font_is_basefont(font)) {
    if (font->encoding_id >= 0) {
      Encoding *enc = Encoding_cache_get(font->encoding_id);
      pdf_add_dict(font->fontdict,
		   pdf_new_name("Encoding"),
		   Encoding_get_resource(enc));
    }
    return;
  }

  pdf_add_dict(font->fontdict, 
	       pdf_new_name("FontDescriptor"),
	       pdf_ref_obj (font->descriptor));

  /*
   * Open Type 1 font.
   */
  {
    if (!font->used_chars)
      ERROR("Type1: Unexpected error.");

    fp = MFOPEN(font->filename, FOPEN_RBIN_MODE);
    if (fp == NULL) {
      ERROR("Type1: Could not open Type1 font.");
    }
  }

  GIDMap     = NULL;
  num_glyphs = 0;

  if (font->encoding_id >= 0) {
    enc_vec = NULL;
  } else {
    enc_vec = NEW(256, char *);
    for (cc = 0; cc < 256; cc++)
      enc_vec[cc] = NULL;
  }

  cff = t1_load_font(enc_vec, 0, fp);
  if (!cff) {
    if (enc_vec)
      RELEASE(enc_vec);
    MFCLOSE(fp);
    ERROR("Could not load Type 1 font: %s", font->filename);
    return;
  }
  MFCLOSE(fp);

  /*
   * Encoding related things.
   */
  if (font->encoding_id >= 0) {
    Encoding *enc;

    enc = Encoding_cache_get(font->encoding_id);
    if (Encoding_is_predefined(enc)) {
      pdf_add_dict(font->fontdict,
		   pdf_new_name("Encoding"),
		   pdf_new_name(Encoding_get_name(enc)));
    } else {
#if 0
      /*
       * Gs not working with this.
       */
      pdf_add_dict(font->fontdict,
		   pdf_new_name("Encoding"),
		   Encoding_get_resource(enc));
#endif
      UC_add_ToUnicode_CMap(font->fontdict, enc);
    }
    enc_vec = Encoding_get_encoding(enc);
  } else {
    pdf_obj *tounicode;
    char    *tmpname;
    /*
     * Create enc_vec and ToUnicode CMap for built-in encoding.
     */
    ASSERT(enc_vec);
    for (cc = 0; cc < 256; cc++) {
      if (!font->used_chars[cc] && enc_vec[cc]) {
	RELEASE(enc_vec[cc]);
	enc_vec[cc] = NULL;
      }
    }
    /*
     * The naming convention for embedded ToUnicode is not clear.
     * We use PseudoUniqTag+FontName-UTF16 here. Not -UCS2, it is to-UTF16BE CMap.
     */
    tmpname = NEW(7+strlen(font->fontname)+strlen("-UTF16")+1, char);
    sprintf(tmpname, "%6s+%s-UTF16", font->uniqueID, font->fontname);
    tounicode = UC_make_ToUnicode_CMap(tmpname, enc_vec);
    if (tounicode) {
      pdf_add_dict(font->fontdict,
		   pdf_new_name("ToUnicode"), pdf_ref_obj(tounicode));
      pdf_release_obj(tounicode);
    }
    RELEASE(tmpname);
  }

  if (!font->fontname)
    ERROR("Fontname undefined...");
  else {
    char *tmpname;

    tmpname = NEW(strlen(font->fontname)+8, char);
    sprintf(tmpname, "%6s+%s", font->uniqueID, font->fontname);
    cff_set_name(cff, tmpname);
    RELEASE(tmpname);
  }

  /* defaultWidthX, CapHeight, etc. */
  get_font_attr(font, cff);
  if (cff_dict_known(cff->private[0], "defaultWidthX")) {
    defaultwidth = cff_dict_get(cff->private[0], "defaultWidthX", 0);
  } else {
    defaultwidth = 0.0;
  }
  if (cff_dict_known(cff->private[0], "nominalWidthX")) {
    nominalwidth = cff_dict_get(cff->private[0], "nominalWidthX", 0);
  } else {
    nominalwidth = 0.0;
  }

  /* Create CFF encoding, charset, sort glyphs */
#define MAX_GLYPHS 1024
  GIDMap = NEW(MAX_GLYPHS, card16);
  {
    int  prev, i;
    long gid;

    cff->encoding = NEW(1, cff_encoding);
    cff->encoding->format      = 1;
    cff->encoding->num_entries = 0;
    cff->encoding->data.range1 = NEW(256, cff_range1);
    cff->encoding->num_supps   = 0;
    cff->encoding->supp        = NEW(256, cff_map);

    charset = NEW(1, cff_charsets);
    charset->format = 0;
    charset->num_entries = 0;
    charset->data.glyphs = NEW(MAX_GLYPHS, s_SID);

    gid = cff_glyph_lookup(cff, ".notdef");
    if (gid < 0)
      ERROR("Type 1 font with no \".notdef\" glyph???");
    GIDMap[0] = (card16) gid;
    if (verbose > 2)
      MESG("[glyphs:/.notdef");
    num_glyphs =  1;
    prev  = -2;
    for (cc = 0; cc < 256; cc++) {
      char  *glyph;
      s_SID  sid;

      glyph = enc_vec[cc];
      if (font->used_chars[cc] &&
	  glyph && strcmp(glyph, ".notdef")) {
	gid = cff_glyph_lookup(cff, glyph);
	if (gid < 1 || gid >= cff->cstrings->count) {
	  WARN("Glyph \"%s\" missing in font \"%s\".", glyph, font->ident);
	  font->used_chars[cc] = 0;
	  continue;
	}
	for (i = 0; i < cc; i++) {
	  if (font->used_chars[i] &&
	      enc_vec[i] && !strcmp(enc_vec[i], glyph))
	    break;
	}
	sid = cff_add_string(cff, glyph); /* FIXME */
	if (i != cc) {
	  cff->encoding->supp[cff->encoding->num_supps].code  = i;
	  cff->encoding->supp[cff->encoding->num_supps].glyph = sid;
	  cff->encoding->num_supps += 1;
	} else {
	  GIDMap[num_glyphs] = (card16) gid;
	  charset->data.glyphs[charset->num_entries] = sid;
	  charset->num_entries += 1;
	  if (cc != prev + 1) {
	    cff->encoding->num_entries += 1;
	    cff->encoding->data.range1[cff->encoding->num_entries-1].first  = cc;
	    cff->encoding->data.range1[cff->encoding->num_entries-1].n_left = 0;
	  } else
	    cff->encoding->data.range1[cff->encoding->num_entries-1].n_left += 1;
	  prev = cc;
	  num_glyphs++;
	  if (verbose > 2)
	    MESG("/%s", glyph);
	}
      } else {
	font->used_chars[cc] = 0;
      }
    }
    if (cff->encoding->num_supps > 0)
      cff->encoding->format |= 0x80;
    else {
      RELEASE(cff->encoding->supp); /* FIXME */
      cff->encoding->supp = NULL;
    }
  }

  widths = NEW(cff->cstrings->count, double);
  /*
   * No more string.
   * The Type 1 seac operator may add another glyph but the glyph name of
   * those glyphs are contained in standard string. The String Index will
   * not be modified after here.
   */
  cff_dict_update(cff->topdict, cff);
  cff_dict_update(cff->private[0], cff);
  cff_update_string(cff);

  {
    cff_index *cstring;
    t1_ginfo   gm;
    card16 gid, ogid;
    long   max = 0;
    int    have_seac = 0;

    offset = 0L;
    cstring = cff_new_index(cff->cstrings->count);
    cstring->data = NULL;
    cstring->offset[0] = 1;

    /* The num_glyphs increases if "seac" operators are used. */
    for (gid = 0; gid < num_glyphs; gid++) {
      if (offset + CS_STR_LEN_MAX >= max) {
	max += CS_STR_LEN_MAX*2;
	cstring->data = RENEW(cstring->data, max, card8);
      }
      ogid = GIDMap[gid];
      offset += t1char_convert_charstring(cstring->data + cstring->offset[gid] - 1, CS_STR_LEN_MAX,
					  cff->cstrings->data + cff->cstrings->offset[ogid] - 1,
					  cff->cstrings->offset[ogid+1] - cff->cstrings->offset[ogid],
					  cff->subrs[0], defaultwidth, nominalwidth, &gm);
      cstring->offset[gid+1] = offset + 1;
      if (gm.use_seac) {
	long  comp_gid, i;
	char *comp_name;
	/*
	 * NOTE:
	 *  1. seac.achar and seac.bchar must be contained in the CFF standard string.
	 *  2. Those characters need not to be encoded.
	 *  3. num_glyphs == charsets->num_entries + 1.
	 */
	comp_name = t1_get_standard_glyph(gm.seac.achar);
	comp_gid  = cff_glyph_lookup(cff, comp_name);
	if (comp_gid < 0) {
	  WARN("Accent char \"%s\" not found. Invalid use of \"seac\" operator.", comp_name);
	  continue;
	}
	for (i = 0; i < num_glyphs; i++) {
	  if (GIDMap[i] == comp_gid)
	    break;
	}
	if (i == num_glyphs) {
	  if (verbose > 2)
	    MESG("/%s", comp_name);
	  GIDMap[num_glyphs++] = comp_gid;
	  charset->data.glyphs[charset->num_entries] = cff_get_sid(cff, comp_name);
	  charset->num_entries += 1;
	}

	comp_name = t1_get_standard_glyph(gm.seac.bchar);
	comp_gid  = cff_glyph_lookup(cff, comp_name);
	if (comp_gid < 0) {
	  WARN("Base char \"%s\" not found. Invalid use of \"seac\" operator.", comp_name);
	  continue;
	}
	for (i = 0; i < num_glyphs; i++) {
	  if (GIDMap[i] == comp_gid)
	    break;
	}
	if (i == num_glyphs) {
	  if (verbose > 2)
	    MESG("/%s", comp_name);
	  GIDMap[num_glyphs++] = comp_gid;
	  charset->data.glyphs[charset->num_entries] = cff_get_sid(cff, comp_name);
	  charset->num_entries += 1;
	}
	have_seac = 1;
      }
      widths[gid] = gm.wx;
    }
    cstring->count = num_glyphs;

    cff_release_index(cff->subrs[0]);
    cff->subrs[0] = NULL;
    RELEASE(cff->subrs);
    cff->subrs    = NULL;

    cff_release_index(cff->cstrings);
    cff->cstrings = cstring;

    cff_release_charsets(cff->charsets);
    cff->charsets = charset;
  }
  if (verbose > 2)
    MESG("]");

  add_metrics(font, cff, widths, num_glyphs);

  offset = write_fontfile(font, cff, num_glyphs);
  if (verbose > 1)
    MESG("[%u glyphs][%ld bytes]", num_glyphs, offset);

  cff_close(cff);

  /* Cleanup */
  if (font->encoding_id < 0 && enc_vec) {
    for (cc = 0; cc < 256; cc++) {
      if (enc_vec[cc])
	RELEASE(enc_vec[cc]);
    }
    RELEASE(enc_vec);
  }
  if (widths)
    RELEASE(widths);
  if (GIDMap)
    RELEASE(GIDMap);

  /*
   * Maybe writing Charset is recommended for subsetted font.
   */

  return;
}
