/*  $Header: /home/cvsroot/dvipdfmx/src/unicode.c,v 1.17 2004/08/15 15:37:08 hirata Exp $

    This is dvipdfmx, an eXtended version of dvipdfm by Mark A. Wicks.

    Copyright (C) 2002 by Jin-Hwan Cho and Shunsaku Hirata,
    the dvipdfmx project team <dvipdfmx@project.ktug.or.kr>

    Copyright (C) 1998, 1999 by Mark A. Wicks <mwicks@kettering.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

/*
 * Unicode related:
 *  Conversion between UTF-* and UCS-*.
 *  ToUnicode CMap
 *
 * Normalization?
 */

#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>

#include "system.h"
#include "mem.h"
#include "error.h"
#include "dpxutil.h"

#include "pdfobj.h"
#include "pdfresource.h"

#include "encodings.h"
#include "agl.h"
#include "unicode.h"

static int __verbose = 0;

#define UC_DEBUG     3
#define UC_DEBUG_STR "UC"

void
UC_set_verbose (void)
{
  __verbose++;
}

#define UC_REPLACEMENT_CHAR 0x0000FFFDL

#define UC_SUR_SHIFT      10
#define UC_SUR_MASK       0x3FFUL
#define UC_SUR_LOW_START  0xDC00UL
#define UC_SUR_HIGH_START 0xD800UL

int
UC_sput_UTF16BE (long ucv, unsigned char **pp, unsigned char *limptr)
{
  int count = 0;
  unsigned char *p = *pp;

  if (ucv >= 0 && ucv <= 0xFFFF) {
    if (p + 2 >= limptr)
      ERROR("Not enough space available...");
    p[0] = (ucv >> 8) & 0xff;
    p[1] = ucv & 0xff;
    count = 2;
  } else if (ucv >= 0x010000L && ucv <= 0x10FFFFL) {
    unsigned short high, low;

    if (p + 4 >= limptr)
      ERROR("Not enough space available...");
    ucv  -= 0x00010000L;
    high = (ucv >> UC_SUR_SHIFT) + UC_SUR_HIGH_START;
    low  = (ucv &  UC_SUR_MASK)  + UC_SUR_LOW_START;
    p[0] = (high >> 8) & 0xff;
    p[1] = (high & 0xff);
    p[2] = (low >> 8) & 0xff;
    p[3] = (low & 0xff);
    count = 4;
  } else {
    if (p + 2 >= limptr)
      ERROR("Not enough space available...");
    p[0] = (UC_REPLACEMENT_CHAR >> 8) & 0xff;
    p[1] = (UC_REPLACEMENT_CHAR & 0xff);
    count = 2;
  }

  *pp += count;
  return count;
}

int
UC_is_valid (long ucv)
{
  if (ucv < 0 || (ucv >= 0x0000D800L && ucv <= 0x0000DFFFL))
    return 0;
  return 1;
}

/* CSI_UNICODE */
#include "cid.h"

#include "cmap.h"
#include "cmap_write.h"

pdf_obj *
UC_make_ToUnicode_CMap (char *cmap_name, char **enc_vec)
{
  pdf_obj  *stream = NULL;
  CMap     *cmap;
  int       code, glyph_count, total_fail_count;
  static unsigned char range_min[1] = {0x00u};
  static unsigned char range_max[1] = {0xFFu};
#define WBUF_SIZE 1024
  unsigned char wbuf[WBUF_SIZE], *p, *endptr;

  ASSERT(cmap_name && enc_vec);

  cmap = CMap_new();
  CMap_set_name (cmap, cmap_name);
  CMap_set_type (cmap, CMAP_TYPE_TO_UNICODE);
  CMap_set_wmode(cmap, 0);

  CMap_set_CIDSysInfo(cmap, &CSI_UNICODE);

  CMap_add_codespacerange(cmap, range_min, range_max, 1);

  glyph_count = total_fail_count = 0;

  endptr = wbuf + WBUF_SIZE;
  for (code = 0; code < 256; code++) {
    if (enc_vec[code] && strcmp(enc_vec[code], ".notdef")) {
      long len;
      int  fail_count;

      wbuf[0] = (code & 0xff);
      p = wbuf + 1;
      len = agl_sput_UTF16BE(enc_vec[code], &p, endptr, &fail_count);
      if (len > 0 && fail_count == 0) {
	CMap_add_bfchar(cmap, wbuf, 1, wbuf+1, len);
      } else {
	/* Conversion of one or more glyph failed, mapping is not
	 * valid, backup.
	 */
	total_fail_count += fail_count;
      }
      glyph_count++;
    }
  }

  if (total_fail_count < (glyph_count + 9)/ 10)
    stream = CMap_create_stream(cmap, 0);
  else {
    if (__verbose) {
      WARN("%d (%d) glyphs missing Unicode mapping...",
	   total_fail_count, glyph_count);
      WARN("ToUnicode CMap \"%s\" removed.", cmap_name);
    }
  }

  CMap_release(cmap);

  return stream;
}

void
UC_add_ToUnicode_CMap (pdf_obj *fontdict, Encoding *encoding)
{
  char    *encoding_name, **encoding_vec;
  char    *cmap_name;
  pdf_obj *cmap_ref;

  ASSERT(fontdict);
  ASSERT(encoding);

  if (Encoding_is_predefined(encoding))
    return;

  encoding_name = Encoding_get_name(encoding);
  encoding_vec  = Encoding_get_encoding(encoding);
  ASSERT(encoding_name && encoding_vec);

  cmap_name = NEW(strlen(encoding_name)+strlen("-UCS2")+1, char);
  strcpy(cmap_name, encoding_name);
  strcat(cmap_name, "-UCS2");

  if (!PDF_findresource(cmap_name, PDF_RES_TYPE_CMAP)) {
    pdf_obj *cmap;

    cmap = UC_make_ToUnicode_CMap(cmap_name, encoding_vec);
    if (cmap)
      PDF_defineresource(cmap_name, cmap, PDF_RES_TYPE_CMAP);
  }

  cmap_ref = PDF_findresource(cmap_name, PDF_RES_TYPE_CMAP);
  if (cmap_ref) {
    pdf_add_dict(fontdict, pdf_new_name("ToUnicode"), cmap_ref);
  }

  RELEASE(cmap_name);

  return;
}
